class SizePolicy:
    expand          : ...
    fixed           : ...

class Orientation:
    vertical        : ...
    horizontal      : ...

class Alignment:
    center          : ...
    vcenter         : ...
    hcenter         : ...
    justify         : ...
    top             : ...
    bottom          : ...
    left            : ...
    right           : ...
    
class Position:
    top             : ...
    bottom          : ...
    left            : ...
    right           : ...

class WidgetType:
    normal          : ...
    static          : ...

class InputType:
    standard        : ...
    password        : ...
    noecho          : ...

class SelectionMode:
    extended        : ...
    noselection     : ...
    multi           : ...
    single          : ...

class WindowStyle:
    windows         : ...
    fusion          : ...
    breeze          : ...
    oxygen          : ...
    curve           : ...

class ButtonStyle:
    circle          : ...
    square          : ...

class OpenFileType:
    single          : ...
    multi           : ...

def call(func: function, *args, **kwargs): ...
def r(list: list) -> str: ...
def editor(): ...